import sys
import getopt
import os
import subprocess as sp

from bare_estate import __version__
import bare_estate.commands as cmd
from bare_estate.config import Configs


SHORT_OPTS = "vnc:r:d:t:C:"
LONG_OPTS = [
    "version",
    "dry-run",
    "config=",
    "repo=",
    "git-dir=",
    "work-tree=",
]


def main(argv=sys.argv, /):
    opts, args = getopt.getopt(argv[1:], SHORT_OPTS, LONG_OPTS)
    status = 0

    if len(args) > 0:
        cmd_name = args[0]
    else:
        cmd_name = "estate"

    dry_run = False
    config_file = None
    options = {}
    for flag, value in opts:
        if flag in ("-v", "--version") and __version__ is None:
            return 1
        elif flag in ("-v", "--version"):
            print("bare-estate version", __version__)
            return 0
        elif flag in ("-n", "--dry-run"):
            dry_run = True
        elif flag in ("-c", "--config"):
            config_file = value
        elif flag in ("-r", "--repo"):
            options["name"] = value
        elif flag in ("-d", "--git-dir"):
            options["git_dir"] = value
        elif flag in ("-t", "--work-tree"):
            options["work_tree"] = value
        elif flag == "-C":
            options["base_dir"] = value

    config = Configs(config_file)
    command = cmd.Command(config)
    run = command.dry_run if dry_run else command.run
    config_file_exists = os.access(command.configs.filename, os.F_OK)

    def update_configs(configs: Configs):
        if config_file_exists:
            with open(configs.filename, mode="w") as file:
                configs.dump(file)
        else:
            configs.create_config_file()

    try:
        if cmd_name == "estate":
            run("status", args)
        elif cmd_name == "init" and dry_run:
            run("init", args)
            print("TOML configuration written to file:", config.filename)
            command.configs.dump(sys.stdout)
        elif cmd_name == "init":
            run("init", args)
            update_configs(command.configs)
        elif cmd_name == "clone" and dry_run:
            run("clone", args)
            print("TOML configuration written to file:", config.filename)
            command.configs.dump(sys.stdout)
        elif cmd_name == "clone":
            run("clone", args)
            update_configs(command.configs)
        else:
            run(command.git, args, **options)
    except RuntimeError as err:
        print("Error:", err, file=sys.stderr)
        status = 1

    return status



import os
import subprocess as sp
import shutil
import re
import typing as t
import getopt

from bare_estate.config import Configs


whitespace = re.compile(r"\s")

ArgV: t.TypeAlias = list[str]
ParsedArgs: t.TypeAlias = tuple[dict[str, t.Any], ArgV]


class Action(t.Protocol):
    def __call__(self, argv: list[str], env=os.environ, **options):
        pass


def _format_shell_arg(arg: str):
    if len(arg) == 0:
        return "''"

    match = whitespace.search(arg)
    space_index = match.start() if match else -1
    eq_index = arg.find("=")

    if match is None:
        return arg
    elif eq_index == -1:
        return f"'{arg}'"
    elif eq_index > 0 and eq_index < space_index:
        return f"{arg[:eq_index+1]}'{arg[eq_index+1:]}'"

    return f"'{arg}'"


class ArgParser:
    """Parse operators and flags from a list of command-line arguments."""

    @staticmethod
    def init(argv: ArgV) -> ParsedArgs:
        short_opts = "d:u:"
        long_opts = ["git-dir=", "url="]

        options: dict[str, t.Any] = {}
        opts, args = getopt.getopt(argv[1:], short_opts, long_opts)

        for flag, value in opts:
            if flag in ("-d", "--git-dir"):
                options["git_dir"] = value
            elif flag in ("-u", "--url"):
                options["url"] = value

        return options, args

    @staticmethod
    def clone(argv: ArgV) -> ParsedArgs:
        short_opts = "r:d:"
        long_opts = ["repo=", "git-dir="]

        options: dict[str, t.Any] = {}
        opts, args = getopt.getopt(argv[1:], short_opts, long_opts)

        for flag, value in opts:
            if flag in ("-r", "--repo"):
                options["name"] = value
            elif flag in ("-d", "--git-dir"):
                options["git_dir"] = value

        return options, args


class Command:
    """
    A helper class to execute git commands.

    It exposes two main methods in 'run' and 'dry_run'. They both
    receive a callback and the parameters to be passed into it. The
    'run' method will execute the command using the input callback
    while 'dry_run' just prints which commands would be executed,
    as one or more shell commands.

    The callback methods implement commands that you would use in a
    normal git project, such as 'init' and 'clone', or you can also use
    the general case function labeled as 'git'. Each of those also
    receive a callback, called action, that receives a list of strings
    representing a git command. The action callback is subprocess.run if
    the calling method is 'run' and self.parse_shell_cmd if it is
    'dry_run'. That way you can either execute the commands as
    subprocesses or print them to stdout as shell commands.
    """
    configs: Configs

    def __init__(self, configs: Configs):
        """Read configuration options from a Configs object."""
        self.configs = configs

    def run(self, comm, *args, **kwargs) -> int:
        """Execute an estate command."""
        caller = getattr(self, comm) if isinstance(comm, str) else comm

        for process in caller(sp.run, *args, **kwargs):
            if process.returncode != 0:
                return process.returncode

        return 0

    def dry_run(self, comm, argv: ArgV, **kwargs):
        """Print shell commands without executing."""
        try:
            caller = getattr(self, comm) if isinstance(comm, str) else comm
        except AttributeError as err:
            caller = self.git

        for cmd in caller(self.parse_shell_cmd, argv, **kwargs):
            print(*cmd)

    @staticmethod
    def parse_shell_cmd(arg_list: list[str], env=os.environ) -> list[str]:
        """
        Returns a list of arguments as a representation of a proper shell
        command. This will add quotation marks around arguments with any
        whitespace characters in them.
        """

        command = []
        env_list = ["GIT_DIR", "GIT_WORK_TREE"]

        for var in env_list:
            if var in env:
                command.append(f"{var}={_format_shell_arg(env[var])}")

        for arg in arg_list:
            command.append(_format_shell_arg(arg))

        return command

    def init(self, action: Action, argv: ArgV, env=os.environ):
        """Initialize a new git repository."""

        if shutil.which("git") is None:
            raise FileNotFoundError(5, "Command not found", "git")

        options, operators = ArgParser.init(argv)

        name = operators[0]
        configs = self.configs
        base_dir = configs.base_dir

        git_dir = options.get("git_dir", name)
        work_tree = operators[1] if len(operators) > 1 else ""

        configs.create_repo({
            "name": name,
            "git_dir": git_dir,
            "work_tree": work_tree,
        })

        if work_tree == "":
            git_cmd = list(self._generate(name, work_tree=None))
            yield action([*git_cmd, "init"], env=env)
            yield action([*git_cmd, "config", "core.bare", "false"], env=env)
            repo_work_tree = ""
        else:
            git_cmd = list(self._generate(
                name, git_dir=None, work_tree=None
            ))
            repo_git_dir = configs.get(name, "git_dir")
            repo_work_tree = str(configs.get(name, "work_tree"))

            yield action([
                *git_cmd, "init",
                f"--separate-git-dir={repo_git_dir}",
                repo_work_tree
            ], env=env)

        work_tree_path = os.path.join(configs.base_dir, repo_work_tree)
        if os.path.abspath(work_tree_path) == env.get("HOME"):
            show = "no"
        else:
            show = "all"
        git_cmd = list(self._generate(name, work_tree=None))
        yield action([*git_cmd, "config", "status.showUntrackedFiles", show],
                     env=env)

    def clone(self, action: Action, argv: ArgV, env=os.environ):
        if shutil.which("git") is None:
            raise FileNotFoundError(5, "Command not found", "git")

        if len(argv) < 2:
            raise RuntimeError("too few arguments")

        options, operators = ArgParser.clone(argv)

        url = operators[0]
        work_tree = operators[1] if len(operators) > 1 else None
        name = options.get("name", os.path.basename(url).rstrip(".git"))
        git_dir = options.get("git_dir", name)

        table = {"name": name, "git_dir": git_dir}
        if work_tree is not None:
            table["work_tree"] = work_tree

        configs = self.configs
        configs.create_repo(table)

        git_cmd = list(self._generate(name, work_tree=None, git_dir=None))
        git_dir = f"--separate-git-dir={os.path.join(configs.git_dir, name)}"
        base_dir = env.get("HOME")

        if work_tree is None:
            clone_cmd = [*git_cmd, "clone", git_dir, url]
        else:
            clone_cmd = [*git_cmd, "clone", git_dir, url, work_tree]

        yield action(clone_cmd, env=env)

    def status(self, action: Action, argv: ArgV, env=os.environ):
        yield action(["/bin/sh", "-c", "true"], env=env)

    def forget(self, action: Action, argv: ArgV, env=os.environ):
        """Remove files from index without deleting."""
        if shutil.which("git") is None:
            raise FileNotFoundError(5, "Command not found", "git")

        raise NotImplementedError("git rm --cached <files>")

    def git(self, action: Action, argv: ArgV, env=os.environ, **options):
        """
        Run generic git commands. Unlike other functions such as 'init'
        and 'clone', this function doesn't parse options from its argv
        parameter. Therefore, options must be provided as named
        arguments.
        """
        if shutil.which("git") is None:
            raise FileNotFoundError(5, "Command not found", "git")

        name = options.get("name", "")
        configs = self.configs
        base_dir = options.get("base_dir", configs.base_dir)
        git_dir = options.get("git_dir", configs.get(name, "git_dir"))
        work_tree = options.get("work_tree", configs.get(name, "work_tree"))

        if git_dir is None:
            raise RuntimeError("git directory was not provided")

        git_cmd = list(self._generate(
            name, base_dir=base_dir, git_dir=git_dir, work_tree=work_tree
        ))
        yield action([*git_cmd, *argv])

    def _generate(self, name, **options):
        configs = self.configs

        yield "git"

        base_dir = options.get("base_dir", configs.base_dir)
        if base_dir is not None:
            yield "-C"
            yield base_dir

        git_dir = options.get("git_dir", configs.get(name, "git_dir"))
        if git_dir is not None:
            yield f"--git-dir={git_dir}"

        work_tree = options.get("work_tree", configs.get(name, "work_tree"))
        if work_tree is not None and work_tree != "":
            yield f"--work-tree={work_tree}"



import sys
import os
import io
import tomllib
import typing as t

import toml


def join(*paths: str) -> str:
    return os.path.join(*paths).rstrip("/")


def home():
    return os.environ["HOME"]


def data_home():
    return os.environ.get("XDG_DATA_HOME", join(home(), ".local/share"))


def config_home():
    return os.environ.get("XDG_CONFIG_HOME", join(home(), ".config"))


class RepoConfigs:
    name: str
    git_dir: str
    work_tree: str

    def __init__(self, table: dict[str, str]):
        self.name = table["name"]
        if self.name == "":
            raise ValueError("repository name cannot be empty")

        self.git_dir = table.get("git_dir", self.name)
        if self.git_dir == "":
            raise ValueError("git_dir cannot be empty")

        self.work_tree = table.get("work_tree", "")

    def __str__(self):
        name = type(self).__name__
        return ('<%s name="%s" git_dir="%s" work_tree="%s">' %
                (name, self.name, self.git_dir, self.work_tree))

    @classmethod
    def from_name(cls, name: str, git_dir: str, work_tree: str):
        """Create a default repo config from a name."""
        table = {
            "name": name,
            "git_dir": join(git_dir, name),
            "work_tree": join(work_tree, ""),
        }

        return cls(table)


class Configs:
    repos: dict[str, RepoConfigs]
    filename: str

    def __init__(self, config_file=None):
        """Read and parse a TOML configuration file."""
        if config_file is None:
            self.filename = join(config_home(), "bare_estate.toml")
            try:
                with open(self.filename) as file:
                    configs = self._read_toml(file)
            except (FileNotFoundError, PermissionError):
                configs = self.default
        elif isinstance(config_file, str):
            configs = self._read_toml(config_file)
            self.filename = config_file
        else:
            configs = self._read_toml(config_file)
            self.filename = join(config_home(), "bare_estate.toml")

        self._configs = {}
        gen_configs = configs.get("config", {})

        self.base_dir = gen_configs.get("base_dir", home())

        self.git_dir = gen_configs.get(
            "git_dir", join(data_home(), "bare_estate")
        )
        if self.git_dir == "":
            raise ValueError("git_dir cannot be empty")

        self.work_tree = gen_configs.get("work_tree", "")

        self.repos = {}

        for table in configs.get("repo", []):
            if "name" not in table:
                raise TypeError("new repository must have a name")

            repo_config = {
                "name": table["name"],
                "git_dir": table.get("git_dir", table["name"]),
                "work_tree": table.get("work_tree", ""),
            }
            self.repos[table["name"]] = RepoConfigs(repo_config)

    @classmethod
    def from_dict(cls, config_dict: dict[str, t.Any]):
        """Create a new Configs instance from a dictionary."""
        conf = cls()
        default = conf.default
        gen_configs = config_dict.get("config", {})
        conf.base_dir = gen_configs.get("base_dir", default["base_dir"])
        conf.git_dir = gen_configs.get("git_dir", default["git_dir"])
        conf.work_tree = gen_configs.get("work_tree", default["work_tree"])

        repos = config_dict.get("repo", config_dict.get("repos", {}))
        for repo_configs in repos:
            if "name" not in repo_configs:
                raise TypeError("new repository must have a name")

            name = repo_configs["name"]
            repo_configs["git_dir"] = repo_configs.get("git_dir", name)
            repo_configs["work_tree"] = repo_configs.get("work_tree", "")

            conf.create_repo(repo_configs)

        return conf

    def __str__(self):
        name = type(self).__name__
        return ('<%s base_dir="%s" git_dir="%s" work_tree="%s">' %
                (name, self.base_dir, self.git_dir, self.work_tree))

    @property
    def base_dir(self) -> str:
        """
        The directory to be used as a starting point for every git
        command. This directory will be passed as argument to the -C
        option. It is set to the user's home directory by default.
        """
        return self._configs["base_dir"]

    @base_dir.setter
    def base_dir(self, value: str):
        self._configs["base_dir"] = value

    @property
    def git_dir(self) -> str:
        """
        Set the path to a directory that contains all the configuration
        repositories.

        Each repository also has its own git_dir property, which is
        either a path relative to the general git_dir or an absolute
        one.

        By default, the general git_dir property is set to
        ${XDG_DATA_HOME}/bare_estate and each repository will be saved
        in a subdirectory which receives the repository's name.
        """
        return self._configs["git_dir"]

    @git_dir.setter
    def git_dir(self, value: str):
        self._configs["git_dir"] = value

    @property
    def work_tree(self) -> str:
        """
        Set the path to the working tree. It can be set as a path
        relative to the base directory or as an absolute path.

        Each repository can also have its own work_tree property,
        denoting a directory that contains its configuration files. In
        this case, it can either be a path relative to the general
        work_tree property or an absolute one.

        By default, each repository will have its work_tree set to the
        base directory.
        """
        return self._configs["work_tree"]

    @work_tree.setter
    def work_tree(self, value: str):
        self._configs["work_tree"] = value

    @property
    def default(self):
        return {
            "base_dir": home(),
            "git_dir": join(data_home(), "bare_estate"),
            "work_tree": "",
            "repo": [],
        }

    def get(self, name: str, key=None) -> RepoConfigs | str | None:
        if name not in self.repos:
            return None
        elif key is None:
            return self.repos[name]

        return join(getattr(self, key, ""), getattr(self.repos[name], key))

    def dump(self, file: io.TextIOBase):
        """Write TOML configuration into a file-like object."""
        repos = []

        for r in self.repos.values():
            repo: dict[str, str] = {"name": r.name}
            if r.git_dir != r.name:
                repo["git_dir"] = r.git_dir
            if r.work_tree != "":
                repo["work_tree"] = r.work_tree
            repos.append(repo)

        c: dict[str, t.Any] = {"config": {"base_dir": self.base_dir}}
        if self.git_dir != self.default["git_dir"]:
            c["config"]["git_dir"] = self.git_dir
        if self.work_tree != self.default["work_tree"]:
            c["config"]["work_tree"] = self.work_tree

        toml.dump(c, file)
        file.write("\n")

        if len(repos) > 0:
            file.write(toml.dumps({"repo": repos}).strip())
            file.write("\n")

    def create_repo(self, table: dict[str, str]):
        if "name" not in table:
            raise TypeError("new repository must have a name")

        self.repos[table["name"]] = RepoConfigs(table)

    def create_config_file(self):
        os.makedirs(os.path.dirname(self.filename), exist_ok=True)
        with open(self.filename, mode="x") as file:
            self.dump(file)

    def _read_toml(self, input_file):
        if isinstance(input_file, str):
            try:
                with open(input_file, mode="rb") as file:
                    return tomllib.load(file)
            except FileNotFoundError:
                return self.default
            except PermissionError:
                print(f"Warning: The file {input_file} doesn't allow reading",
                      "Using default configuration as fallback",
                      sep="\n",
                      file=sys.stderr)
                return self.default
        elif not isinstance(input_file, io.IOBase):
            raise TypeError("Input must be a file path or a file-like object")
        elif not hasattr(input_file, "read"):
            raise ValueError("File must be readable")
        elif isinstance(input_file, io.BufferedIOBase):
            return tomllib.load(input_file)
        elif isinstance(input_file, io.TextIOBase):
            return toml.load(input_file)
        else:
            raise ValueError("Unsupported file-like object type")
